%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(ddll_SUITE).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Checks if the dynamic driver and linker loader works.
%%%
%%% These tests can only be run installed (outside clearcase).
%%%
%%% XXX In this suite is missing test cases for reference counts
%%% and that drivers are unloaded when their processes die.
%%% (For me to add :-)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


-export([all/1, ddll_test/1, errors/1,
	 autostart/1, reference_count/1,
	 kill_port/1]).
% Private exports
-export([autostart_test1/0, autostart_test2/0,
	 echo_loader/2, properties/1, load_and_unload/1]).

-import(ordsets,[set_to_list/1,list_to_set/1,subtract/2]).

-include("test_server.hrl").

all(suite) -> {req, dynamic_loading, [ddll_test, errors,
				      autostart,
				      reference_count,
				      kill_port,
				      properties,
				      load_and_unload
				     ]}.

ddll_test(suite) -> [];
ddll_test(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line Path = ?config(data_dir, Config),

    ?line erl_ddll:start(),

    %% Load the echo driver and verify that it was loaded.
    {ok,L1,L2}=load_echo_driver(Path),

    %% Verify that the driver works.

    ?line Port = open_port({spawn, echo_drv}, [eof]),
    ?line {hej, "hopp",4711,123445567436543653} = 
	erlang:port_call(Port,{hej, "hopp",4711,123445567436543653}),
    ?line {hej, "hopp",4711,123445567436543653} = 
	erlang:port_call(Port,47,{hej, "hopp",4711,123445567436543653}),
    ?line Port ! {self(), {command, "text"}},
    ?line 1 = receive 
		  {Port, {data, "text"}} -> 1;
		  Other -> 2
	      after 
		  1000 -> 2
	      end,
    ?line Port ! {self(), close},
    ?line receive {Port, closed} -> ok end,

    %% Unload the driver and verify that it was unloaded.
    ok=unload_echo_driver(L1,L2),

    ?line {error, {already_started, _}} = erl_ddll:start(),
    ?line ok = erl_ddll:stop(),

    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Tests errors having to do with bad drivers.

errors(suite) -> [];
errors(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line Path = ?config(data_dir, Config),

    ?line {ok, Pid} = erl_ddll:start(),
    ?line {ok, L1} = erl_ddll:loaded_drivers(),

    ?line {error, {open_error, _}} = erl_ddll:load_driver(Path, bad_name),
    ?line {error, driver_init_failed} = erl_ddll:load_driver(Path, initfail_drv),
    ?line {error, bad_driver_name} = erl_ddll:load_driver(Path, wrongname_drv),

    %% We assume that there is a statically linked driver named "ddll":
    ?line {error, linked_in_driver} = erl_ddll:unload_driver(ddll),
    ?line {error, not_loaded} = erl_ddll:unload_driver("__pucko_driver__"),
	
    case os:type() of
	{unix, _} ->
	    ?line {error, no_driver_init} =
		erl_ddll:load_driver(Path, noinit_drv);
	_ ->
	    ok
    end,

    ?line {ok, L1} = erl_ddll:loaded_drivers(),
    ?line ok = erl_ddll:stop(),

    ?line test_server:timetrap_cancel(Dog),
    ok.


autostart(suite) ->
    [];
autostart(doc) ->
    ["Tests that erl_ddll:loaded_drivers() & erl_ddll:stop() ",
     "doesn't hang if the erl_ddll server isn't started. ",
     "Also tests that the ddll server is automaticly started ",
     "when necessary."];
autostart(Config) when list(Config) ->
    % This is kind of tricky, as erl_ddll:stop() can hang
    % when we stop the server, and we have to make sure that
    % we are testing erl_ddll:loaded_drivers() when the server
    % is not running.
    ?line process_flag(trap_exit, true),
    ?line Pid1=spawn_link(?MODULE, autostart_test1, []),
    ?line autostart_test_hang(Pid1, "erl_ddll:stop() hangs"),
    ?line Pid2=spawn_link(?MODULE, autostart_test2, []),
    ?line autostart_test_hang(Pid2, "erl_ddll:loaded_drivers() hangs"),
    ?line process_flag(trap_exit, false),

    % make sure the server really got started.
    ?line {error, {already_started, _}} = erl_ddll:start(),
    ?line ok=erl_ddll:stop(),
    ok.
   
autostart_test1() ->
    erl_ddll:stop().
autostart_test2() ->
    erl_ddll:loaded_drivers().
autostart_test_hang(Pid, Reason) ->
    receive
	{'EXIT', Pid, normal} -> ok;
	{'EXIT', Pid, Error1} -> ok % It is supposed to crash.
    after 3000 ->
	    exit(Pid, kill),
	    test_server:fail(Reason)
    end.



reference_count(suite) ->
    [];
reference_count(doc) ->
    ["Check that drivers are unloaded when their reference count ",
     "reaches zero, and that they cannot be unloaded while ",
     "they are still referenced."];
reference_count(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line Path = ?config(data_dir, Config),

    ?line erl_ddll:start(),

    %% Spawn a process that loads the driver (and holds a reference
    %% to it).
    Pid1=spawn_link(?MODULE, echo_loader, [Path, self()]),
    receive
	{Pid1, echo_loaded} -> ok
    after 2000 -> test_server:fail("echo_loader failed to start.")
    end,

    Pid1 ! {self(), die},
    ?line test_server:sleep(200),   % Give time to unload.
    % Verify that the driver was automaticly unloaded when the
    % process died.
    ?line {error, not_loaded}=erl_ddll:unload_driver(echo_drv),
    
    ?line ok=erl_ddll:stop(),
    ?line test_server:timetrap_cancel(Dog),
    ok.

% Loads the echo driver, send msg to started, sits and waits to
% get a signal to die, then unloads the driver and terminates.
echo_loader(Path, Starter) ->
    ?line {ok, L1, L2}=load_echo_driver(Path),
    ?line Starter ! {self(), echo_loaded},
    receive
	{Starter, die} ->
	    ?line unload_echo_driver(L1,L2)
    end.


kill_port(doc) ->
    ["Test that a port that uses a driver is killed when the ",
     "process that loaded the driver dies."];
kill_port(suite) ->
    [];
kill_port(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line Path = ?config(data_dir, Config),

    ?line erl_ddll:start(),

    %% Spawn a process that loads the driver (and holds a reference
    %% to it).
    ?line Pid1=spawn(?MODULE, echo_loader, [Path, self()]),
    ?line receive
	      {Pid1, echo_loaded} ->
		  ok
	  after 3000 ->
		  ?line exit(Pid1, kill),
		  ?line test_server:fail("echo_loader failed to start.")
	  end,

    % Spawn off a port that uses the driver.
    ?line Port = open_port({spawn, echo_drv}, [eof]),

    % Kill the process / unload the driver.
    ?line process_flag(trap_exit, true),
    ?line exit(Pid1, kill),
    ?line test_server:sleep(200),    % Give some time to unload.
    ?line {error, not_loaded} = erl_ddll:unload_driver(echo_drv),

    % See if the port is killed.
    receive
	{'EXIT', Port, Reason} ->
	    io:format("Port exited with reason ~w", [Reason])
    after 5000 ->
	    ?line test_server:fail("Echo port did not terminate.")
    end,

    %% Cleanup and exit.
    ?line ok=erl_ddll:stop(),
    ?line test_server:timetrap_cancel(Dog),
    ok.

properties(suite) -> [];
properties(doc) -> ["Test that a process that loaded a driver ",
		    "is the only process that can unload it."];
properties(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line Path = ?config(data_dir, Config),

    ?line erl_ddll:start(),

    % Let another process load the echo driver.
    Pid=spawn_link(?MODULE, echo_loader, [Path, self()]),
    receive
	{Pid, echo_loaded} -> ok
    after 2000 -> test_server:fail("echo_loader failed to start.")
    end,

    % Try to unload the driver from this process (the wrong one).
    ?line {error, _} = erl_ddll:unload_driver(echo_drv),
    ?line {ok, Drivers} = erl_ddll:loaded_drivers(),
    ?line case lists:member("echo_drv", Drivers) of
	      true ->
		  ok;
	      false ->
		  test_server:fail("Unload from wrong process "
				   "succeeded.")
	  end,

    % Unload the driver and terminate dummy process.
    ?line Pid ! {self(), die},
    ?line test_server:sleep(200),   % Give time to unload.
    ?line ok=erl_ddll:stop(),
    ?line test_server:timetrap_cancel(Dog),
    ok.

load_and_unload(suite) -> [];
load_and_unload(doc) -> ["Load two drivers and unload them in load order."];
load_and_unload(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(60)),
    ?line Path = ?config(data_dir, Config),
    ?line {ok, Pid} = erl_ddll:start(),
    ?line {ok, Loaded_drivers1} = erl_ddll:loaded_drivers(),
    ?line ok = erl_ddll:load_driver(Path, echo_drv),
    ?line ok = erl_ddll:load_driver(Path, dummy_drv),
    ?line ok = erl_ddll:unload_driver(echo_drv),
    ?line ok = erl_ddll:unload_driver(dummy_drv),
    ?line {ok, Loaded_drivers2} = erl_ddll:loaded_drivers(),
    ?line Set1 = ordsets:list_to_set(Loaded_drivers1),
    ?line Set2 = ordsets:list_to_set(Loaded_drivers2),
    ?line io:format("~p == ~p\n", [Loaded_drivers1, Loaded_drivers2]),
    ?line [] == ordsets:set_to_list(ordsets:subtract(Set2, Set1)),
    ?line ok = erl_ddll:stop(),

    ?line test_server:timetrap_cancel(Dog),
    ok.    

% Load and unload the echo_drv driver.
% Make sure that the driver doesn't exist before we load it,
% and that it exists before we unload it.
load_echo_driver(Path) ->
    ?line {ok, L1} = erl_ddll:loaded_drivers(),
    ?line ok = erl_ddll:load_driver(Path, echo_drv),
    ?line {ok, L2} = erl_ddll:loaded_drivers(),
    ?line ["echo_drv"] = set_to_list(subtract(list_to_set(L2), 
					      list_to_set(L1))),
    {ok,L1,L2}.

unload_echo_driver(L1,L2) ->
    ?line {ok, L2} = erl_ddll:loaded_drivers(),
    ?line ok = erl_ddll:unload_driver(echo_drv),
    ?line {ok, L3} = erl_ddll:loaded_drivers(),
    ?line [] = set_to_list(subtract(list_to_set(L3),
				    list_to_set(L1))),
    ok.

