%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% Define to run outside of test server
%%%
%%% -define(STANDALONE,1).
%%%
%%%
%%% Define for debug output
%%%
%%% -define(debug,1).

-module(trace_meta_SUITE).

%% Exported end user tests
-export([basic_test/0, return_test/0, on_and_off_test/0, stack_grow_test/0,
	 info_test/0, tracer_test/0, combo_test/0]).

%% Internal exports
-export([exported/1, exported_wrap/1, loop/4, receiver/1]).

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Test server related stuff
%%

-ifdef(STANDALONE).
-define(config(A,B),config(A,B)).
-export([config/2]).
-else.
-include("test_server.hrl").
-endif.
 
-ifdef(debug).
-ifdef(STANDALONE).
-define(line, erlang:display({?MODULE,?LINE}), ).
-endif.
-define(dbgformat(A,B),io:format(A,B)).
-else.
-ifdef(STANDALONE).
-define(line, noop, ).
-endif.
-define(dbgformat(A,B),noop).
-endif.
 
-ifdef(STANDALONE).
config(priv_dir,_) ->
    ".".
-else.
%% When run in test server.
-export([all/1, init_per_testcase/2, fin_per_testcase/2, not_run/1]).
-export([basic/1, return/1, on_and_off/1, stack_grow/1, 
	 info/1, tracer/1, combo/1]).
	 
init_per_testcase(Case, Config) ->
    ?line Dog=test_server:timetrap(test_server:seconds(30)),
    [{watchdog, Dog}|Config].

fin_per_testcase(Case, Config) ->
    shutdown(),
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.
all(doc) ->
    ["Test meta tracing of local function calls and return trace."];
all(suite) ->
    case test_server:is_native(?MODULE) of
	true -> [not_run];
	false -> [basic, return, on_and_off, stack_grow, 
		  info, tracer, combo]
    end.

not_run(Config) when list(Config) -> 
    {skipped,"Native code"}.

basic(suite) ->
    [];
basic(doc) ->
    ["Tests basic meta trace"];
basic(Config) when list(Config) ->
    basic_test().

return(suite) ->
    [];
return(doc) ->
    ["Tests return trace"];
return(Config) when list(Config) ->
    return_test().
 
on_and_off(suite) ->
    [];
on_and_off(doc) ->
    ["Tests turning trace parameters on and off"];
on_and_off(Config) when list(Config) ->
    on_and_off_test().
 
stack_grow(suite) ->
    [];
stack_grow(doc) ->
    ["Tests the stack growth during return traces"];
stack_grow(Config) when list(Config) ->
    stack_grow_test().
 
info(suite) ->
    [];
info(doc) ->
    ["Tests the trace_info BIF"];
info(Config) when list(Config) ->
    info_test().
 
tracer(suite) ->
    [];
tracer(doc) ->
    ["Tests stopping and changing tracer process"];
tracer(Config) when list(Config) ->
    tracer_test().

combo(suite) ->
    [];
combo(doc) ->
    ["Tests combining local call trace with meta trace"];
combo(Config) when list(Config) ->
    combo_test().

-endif.

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Result examination macros

-define(CT(P,MFA),{trace,P,call,MFA}).
-define(CTT(P, MFA),{trace_ts,P,call,MFA,{_,_,_}}).
-define(RF(P,MFA,V),{trace,P,return_from,MFA,V}).
-define(RFT(P,MFA,V),{trace_ts,P,return_from,MFA,V,{_,_,_}}).
-define(RT(P,MFA),{trace,P,return_to,MFA}).
-define(RTT(P,MFA),{trace_ts,P,return_to,MFA,{_,_,_}}).

-define(EXPECT(List), (
		fun() -> 
			case collect() 
			    of 
			    List ->
				ok;
			    Else ->
				exit({unexpected_result, Else, 
				      {?MODULE,?LINE}})
			end
		end)()).

%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The Tests
%%%

basic_test() ->
    ?line Pid = setup(),
    ?line erlang:trace_pattern({?MODULE,'_','_'},[],[meta]),
    ?line [1,1,1,0] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CTT(Pid,{?MODULE,exported_wrap,[1]}), 
		   ?CTT(Pid,{?MODULE,exported,[1]}),
		   ?CTT(Pid,{?MODULE,local,[1]}),
		   ?CTT(Pid,{?MODULE,local2,[1]}),
		   ?CTT(Pid,{?MODULE,local_tail,[1]})]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},false,[meta]),
    ?line [1,1,1,0] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([]),
    ?line [1,1,1,0] = lambda_slave(fun() ->
					   exported_wrap(1)
				   end),
    ?line ?EXPECT([]), 
    ?line erlang:trace_pattern({?MODULE,'_','_'},[],[meta]),
    ?line [1,1,1,0] = lambda_slave(fun() ->
					   exported_wrap(1)
				   end),
    ?line ?EXPECT([?CTT(Pid,{?MODULE,_,_}), %% The fun
		   ?CTT(Pid,{?MODULE,exported_wrap,[1]}), 
		   ?CTT(Pid,{?MODULE,exported,[1]}),
		   ?CTT(Pid,{?MODULE,local,[1]}),
		   ?CTT(Pid,{?MODULE,local2,[1]}),
		   ?CTT(Pid,{?MODULE,local_tail,[1]})]),
    ?line erlang:trace_pattern({?MODULE,'_','_'},false,[meta]),
    ?line shutdown(),
    ok.

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

return_test() ->
    ?line Pid = setup(),
    ?line erlang:trace_pattern({?MODULE,'_','_'},[{'_',[],[{return_trace}]}],
			 [meta]),
    ?line erlang:trace_pattern({erlang,phash2,'_'},[{'_',[],[{return_trace}]}],
			 [meta]),
    ?line [1,1,1,0] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CTT(Pid,{?MODULE,exported_wrap,[1]}), 
		   ?CTT(Pid,{?MODULE,exported,[1]}),
		   ?CTT(Pid,{?MODULE,local,[1]}),
		   ?CTT(Pid,{?MODULE,local2,[1]}),
		   ?CTT(Pid,{?MODULE,local_tail,[1]}),
		   ?CTT(Pid,{erlang,phash2,[1,1]}),
		   ?RFT(Pid,{erlang,phash2,2},0),
		   ?RFT(Pid,{?MODULE,local_tail,1},[1,0]),
		   ?RFT(Pid,{?MODULE,local2,1},[1,0]),
		   ?RFT(Pid,{?MODULE,local,1},[1,1,0]),
		   ?RFT(Pid,{?MODULE,exported,1},[1,1,1,0]),
		   ?RFT(Pid,{?MODULE,exported_wrap,1},[1,1,1,0])]),
    ?line shutdown(),
    ok.

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

on_and_off_test() ->
    ?line Pid = setup(),
    ?line 1 = erlang:trace_pattern({?MODULE,local_tail,1},[],[meta]),
    ?line LocalTail = fun() ->
			      local_tail(1)
		      end,
    ?line [1,0] = lambda_slave(LocalTail),
    ?line ?EXPECT([?CTT(Pid,{?MODULE,local_tail,[1]})]),
    ?line 0 = erlang:trace_pattern({?MODULE,local_tail,1},[],[global]),
    ?line [1,0] = lambda_slave(LocalTail),
    ?line ?EXPECT([]),
    ?line 1 = erlang:trace_pattern({?MODULE,exported_wrap,1},[],[meta]),
    ?line [1,1,1,0] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CTT(Pid,{?MODULE,exported_wrap,[1]})]),
    ?line 1 = erlang:trace_pattern({erlang,phash2,2},[],[meta]),
    ?line [1,1,1,0] = apply_slave(?MODULE,exported_wrap,[1]),
    ?line ?EXPECT([?CTT(Pid,{?MODULE,exported_wrap,[1]}),
		   ?CTT(Pid,{erlang,phash2,[1,1]})]),
    ?line shutdown(),
    ?line erlang:trace_pattern({'_','_','_'},false,[meta]),
    ?line N = erlang:trace_pattern({erlang,'_','_'},true,[meta]),
    ?line case erlang:trace_pattern({erlang,'_','_'},false,[meta]) of
	      N -> 
		  ok;
	      Else ->
		  exit({number_missmatch, {expected, N}, {got, Else}})
	  end,
    ?line case erlang:trace_pattern({erlang,'_','_'},false,[meta]) of
	      N -> 
		  ok;
	      Else2 ->
		  exit({number_missmatch, {expected, N}, {got, Else2}})
	  end,
    ok.
    
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stack_grow_test() ->    
    ?line Pid = setup(),
    ?line 1 = erlang:trace_pattern({?MODULE,loop,4},
				   [{'_',[],[{return_trace}]}],[meta]),
    ?line Num = 1 bsl 16,
    ?line Surface =
	fun (This, [?RFT(P,{?MODULE,loop,4},N) | L], N) when P == Pid->
		if N == Num ->
			L;
		   true ->
			This(This, L, N+1)
		end
	end,
    ?line Dive = 
	fun (This, [?CTT(P,{?MODULE,loop,[{hej,hopp},[a,b,c],4.5,N]}) 
		    | L], N) when P == Pid->
		if N == 0 ->
			Surface(Surface, L, 0);
		   true ->
			This(This, L, N-1)
		end
	end,
    ?line apply_slave(?MODULE,loop,[{hej,hopp},[a,b,c],4.5,Num]),
    ?line List = collect(),
%     ?line apply_slave_async(?MODULE,loop,[{hej,hopp},[a,b,c],4.5,Num]),
%     ?line List = collect(test_server:seconds(5)),
    ?line [] = Dive(Dive, List, Num),
    ok.

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

info_test() ->
    ?line setup(),
    ?line Prog = [{['$1'],[{is_integer,'$1'}],[{message, false}]},
		  {'_',[],[]}],
    ?line Self = self(),
    ?line GoOn = make_ref(),
    
    ?line Pid = 
	spawn_link(
	  fun () ->
		  erlang:trace_pattern({?MODULE,exported_wrap,1}, 
				       Prog, [{meta, Self}]),
		  Self ! {self(), GoOn}
	  end),
    ?line receive {Pid, GoOn} -> ok end,
    ?line {traced,false} = erlang:trace_info({?MODULE,exported_wrap,1}, traced),
    ?line {match_spec, false} = 
	erlang:trace_info({?MODULE,exported_wrap,1}, match_spec),
    ?line {meta, Self} = erlang:trace_info({?MODULE,exported_wrap,1}, meta),
    ?line {meta_match_spec, MMS} = 
	erlang:trace_info({?MODULE,exported_wrap,1}, meta_match_spec),
    ?line case MMS of
	      Prog ->
		  ok;
	      Wrong ->
		  exit({bad_result, {erlang,trace_info,
				     [{?MODULE,exported_wrap,1},
				      meta_match_spec]},
			{expected, Prog}, {got, Wrong}})
	  end,
    ?line erlang:garbage_collect(self()),
    ?line receive
	  after 1 ->
		  ok
	  end,
    ?line io:format("~p~n",[MMS]),
    ?line {meta_match_spec,MMS2} = 
	erlang:trace_info({?MODULE,exported_wrap,1}, meta_match_spec),
    ?line io:format("~p~n",[MMS2]),
    ?line case MMS2 of
	      Prog ->
		  ok;
	      Wrong2 ->
		  exit({bad_result, {erlang,trace_info,
				     [{?MODULE,exported_wrap,1},
				      meta_match_spec]},
			{expected, Prog}, {got, Wrong2}})
	  end,
    ?line {all, [_|_]=L} = erlang:trace_info({?MODULE,exported_wrap,1}, all),
    ?line {value, {meta, Self}} = 
	lists:keysearch(meta, 1, L),
    ?line {value, {meta_match_spec, MMS}} = 
	lists:keysearch(meta_match_spec, 1, L),
    
    ?line erlang:trace_pattern({?MODULE,exported_wrap,1}, true, [meta]),
    ?line {meta_match_spec, []} = 
	erlang:trace_info({?MODULE,exported_wrap,1}, meta_match_spec),
    
    ?line erlang:trace_pattern({?MODULE,exported_wrap,1}, false, [meta]),
    ?line {meta, false} = erlang:trace_info({?MODULE,exported_wrap,1}, meta),
    ?line {meta_match_spec, false} = 
	erlang:trace_info({?MODULE,exported_wrap,1}, meta_match_spec),
    ?line {all, false} = erlang:trace_info({?MODULE,exported_wrap,1}, all),
    
    ?line shutdown(),
    ok.

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tracer_test() ->
    ?line Slave = setup(),
    ?line Self = self(),
    
    ?line MatchSpec = [{'_',[],[{return_trace}]}],
    ?line Tracer1 = spawn_link(fun () -> relay_n(3, Self) end),
    ?line Setter = 
	spawn_link(
	  fun () ->
		  erlang:trace_pattern({?MODULE,receiver,1},
				       MatchSpec,
				       [{meta,Tracer1}]),
		  erlang:trace_pattern({erlang,phash2,2},
				       MatchSpec,
				       [{meta,Tracer1}]),
		  Self ! {self(), done}
	  end),
    ?line receive {Setter, done} -> ok end,
    ?line Ref = make_ref(),
    ?line apply_slave_async(?MODULE, receiver, [Ref]),
    ?line [{Tracer1,?CTT(Slave,{?MODULE,receiver,[Ref]})},
	   {Tracer1,?CTT(Slave,{erlang,phash2,[1,1]})},
	   {Tracer1,?RFT(Slave,{erlang,phash2,2},0)}] = collect(100),
    %% Initiate a return_trace that will fail since the tracer just stopped
    ?line Slave ! Ref, 
    ?line [] = collect(100),
    %% The breakpoint has not been hit since the tracer stopped
    ?line {meta,Tracer1} = 
	erlang:trace_info({?MODULE,receiver,1}, meta),
    ?line {meta_match_spec, MatchSpec} = 
	erlang:trace_info({?MODULE,receiver,1}, meta_match_spec),
    ?line {meta,Tracer1} = 
	erlang:trace_info({erlang,phash2,2}, meta),
    ?line {meta_match_spec, MatchSpec} = 
	erlang:trace_info({erlang,phash2,2}, meta_match_spec),
    %% Initiate trace messages that will fail
    ?line Ref2 = make_ref(),
    ?line apply_slave_async(?MODULE, receiver, [Ref2]),
    ?line Slave ! Ref2,
    ?line [] = collect(100),
    ?line {meta,[]} = 
	erlang:trace_info({?MODULE,receiver,1}, meta),
    ?line {meta_match_spec, MatchSpec} = 
	erlang:trace_info({?MODULE,receiver,1}, meta_match_spec),
    ?line {meta,[]} =
	erlang:trace_info({erlang,phash2,2}, meta),
    ?line {meta_match_spec, MatchSpec} = 
	erlang:trace_info({erlang,phash2,2}, meta_match_spec),
    %% Change tracer
    ?line Tracer2 = spawn_link(fun () -> relay_n(4, Self) end),
    ?line erlang:trace_pattern({?MODULE,receiver,1}, 
			       MatchSpec, 
			       [{meta,Tracer2}]),
    ?line erlang:trace_pattern({erlang,phash2,2},
			       MatchSpec,
			       [{meta,Tracer2}]),
    ?line Ref3 = make_ref(),
    ?line apply_slave_async(?MODULE, receiver, [Ref3]),
    ?line [{Tracer2,?CTT(Slave,{?MODULE,receiver,[Ref3]})},
	   {Tracer2,?CTT(Slave,{erlang,phash2,[1,1]})},
	   {Tracer2,?RFT(Slave,{erlang,phash2,2},0)}] = collect(100),
    %% Change tracer between call trace and return trace
    ?line Tracer3 = spawn_link(fun () -> relay_n(4, Self) end),
    ?line erlang:trace_pattern({?MODULE,receiver,1}, 
			       MatchSpec, 
			       [{meta,Tracer3}]),
    ?line erlang:trace_pattern({erlang,phash2,2},
			       MatchSpec,
			       [{meta,Tracer3}]),
    ?line Slave ! Ref3, 
    %% The return trace should still come from Tracer2
    ?line [{Tracer2,?RFT(Slave,{?MODULE,receiver,1},Ref3)}] = collect(100),
    ?line Ref4 = make_ref(),
    %% Now should Tracer3 be used
    ?line apply_slave_async(?MODULE, receiver, [Ref4]),
    ?line Slave ! Ref4,
    ?line [{Tracer3,?CTT(Slave,{?MODULE,receiver,[Ref4]})},
	   {Tracer3,?CTT(Slave,{erlang,phash2,[1,1]})},
	   {Tracer3,?RFT(Slave,{erlang,phash2,2},0)},
	   {Tracer3,?RFT(Slave,{?MODULE,receiver,1},Ref4)}] = collect(100),
    %% The breakpoint has not been hit since the tracer stopped
    ?line {meta,Tracer3} = 
	erlang:trace_info({?MODULE,receiver,1}, meta),
    ?line {meta_match_spec, MatchSpec} = 
	erlang:trace_info({?MODULE,receiver,1}, meta_match_spec),
    ?line {meta,Tracer3} = 
	erlang:trace_info({erlang,phash2,2}, meta),
    ?line {meta_match_spec, MatchSpec} = 
	erlang:trace_info({erlang,phash2,2}, meta_match_spec),
    
    ?line shutdown(),
    ok.

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

combo_test() ->
    ?line Slave = setup(),
    ?line Self = self(),
    
    ?line MatchSpec = [{'_',[],[{return_trace}]}],
    ?line Flags = lists:sort([call, return_to]),
    ?line LocalTracer = spawn_link(fun () -> relay_n(6, Self) end),
    ?line MetaTracer = spawn_link(fun () -> relay_n(4, Self) end),
    ?line 1 = erlang:trace_pattern({?MODULE,receiver,1}, 
				   MatchSpec,
				   [local,{meta,MetaTracer}]),
    ?line 1 = erlang:trace_pattern({erlang,phash2,2}, 
				   MatchSpec,
				   [local,{meta,MetaTracer}]),
    ?line 1 = erlang:trace(Slave, true, 
			   [{tracer,LocalTracer} | Flags]),
    %%
    ?line {all, TraceInfo1} = 
	erlang:trace_info({?MODULE,receiver,1}, all),
    ?line {meta,MetaTracer} = 
	erlang:trace_info({?MODULE,receiver,1}, meta),
    ?line {value,{meta,MetaTracer}} = 
	lists:keysearch(meta, 1, TraceInfo1),
    ?line {meta_match_spec,MatchSpec} = 
	erlang:trace_info({?MODULE,receiver,1}, meta_match_spec),
    ?line {value,{meta_match_spec,MatchSpec}} = 
	lists:keysearch(meta_match_spec, 1, TraceInfo1),
    ?line {traced,local} = 
	erlang:trace_info({?MODULE,receiver,1}, traced),
    ?line {value,{traced,local}} = 
	lists:keysearch(traced, 1, TraceInfo1),
    ?line {match_spec,MatchSpec} = 
	erlang:trace_info({?MODULE,receiver,1}, match_spec),
    ?line {value,{match_spec,MatchSpec}} = 
	lists:keysearch(match_spec, 1, TraceInfo1),
    %%
    ?line {all, TraceInfo2} = 
	erlang:trace_info({erlang,phash2,2}, all),
    ?line {meta,MetaTracer} = 
	erlang:trace_info({erlang,phash2,2}, meta),
    ?line {value,{meta,MetaTracer}} = 
	lists:keysearch(meta, 1, TraceInfo2),
    ?line {meta_match_spec,MatchSpec} = 
	erlang:trace_info({erlang,phash2,2}, meta_match_spec),
    ?line {value,{meta_match_spec,MatchSpec}} = 
	lists:keysearch(meta_match_spec, 1, TraceInfo2),
    ?line {traced,local} = 
	erlang:trace_info({erlang,phash2,2}, traced),
    ?line {value,{traced,local}} = 
	lists:keysearch(traced, 1, TraceInfo2),
    ?line {match_spec,MatchSpec} = 
	erlang:trace_info({erlang,phash2,2}, match_spec),
    ?line {value,{match_spec,MatchSpec}} = 
	lists:keysearch(match_spec, 1, TraceInfo2),
    %%
    ?line {flags,Flags1} = erlang:trace_info(Slave, flags),
    ?line Flags = lists:sort(Flags1),
    ?line {tracer,LocalTracer} = erlang:trace_info(Slave, tracer),
    %%
    ?line Ref = make_ref(),
    ?line apply_slave_async(?MODULE, receiver, [Ref]),
    ?line Slave ! Ref,
    ?line List = collect(100),
    ?line {MetaR, LocalR} = 
	lists:foldl(
	  fun ({P,X}, {M,L}) when P == MetaTracer ->
		  {[X|M],L};
	      ({P,X}, {M,L}) when P == LocalTracer ->
		  {M,[X|L]}
	  end,
	  {[],[]},
	  List),
    ?line Meta = lists:reverse(MetaR),
    ?line Local = lists:reverse(LocalR),
    ?line [?CTT(Slave,{?MODULE,receiver,[Ref]}),
	   ?CTT(Slave,{erlang,phash2,[1,1]}),
	   ?RFT(Slave,{erlang,phash2,2},0),
	   ?RFT(Slave,{?MODULE,receiver,1},Ref)] = Meta,
    ?line [?CT(Slave,{?MODULE,receiver,[Ref]}),
	   ?CT(Slave,{erlang,phash2,[1,1]}) | Local1] = Local,
    ?line Local2 = 
	case Local1 of
	    [?RF(Slave,{erlang,phash2,2},0),
	     ?RT(Slave,{?MODULE,receiver,1}) | L2] ->
		L2;
	    [?RT(Slave,{?MODULE,receiver,1}),
	     ?RF(Slave,{erlang,phash2,2},0) | L2] ->
		L2
	end,
    ?line case Local2 of
	      [?RF(Slave,{?MODULE,receiver,1},Ref),
	       ?RT(Slave,{?MODULE,slave,1})] ->
		  ok;
	      [?RT(Slave,{?MODULE,slave,1}),
	       ?RF(Slave,{?MODULE,receiver,1},Ref)] ->
		  ok
	  end,
    
    ?line shutdown(),
    ok.

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Trace target functions

loop(D1,D2,D3,0) ->
    io:format("~p~n",[[D1,D2,D3]]),
    0;
loop(D1,D2,D3,N) ->
    max(N,loop(D1,D2,D3,N-1)).

max(A,B) when A > B ->
    A;
max(A,B) ->
    B.



exported_wrap(Val) ->
    exported(Val).

exported(Val) ->
    [Val | local(Val)]. %% Non tail recursive local call

local(Val) ->
    [Val | local2(Val)]. %% Non tail recursive local call

local2(Val) ->
    local_tail(Val). %% Tail recursive call

local_tail(Val) ->
    [Val , erlang:phash2(1,1)]. 



receiver(Msg) ->
    erlang:phash2(1,1),
    receive Msg -> Msg end.
	    
%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Trace target process and utilities

slave(Sync) ->
    Sync ! sync,
    receive
	{From,apply, M, F, A} ->
	    ?line ?dbgformat("Apply: ~p:~p/~p (~p)~n",[M,F,length(A),A]),
	    ?line Res = apply(M,F,A),
	    ?line ?dbgformat("done Apply: ~p:~p/~p (~p)~n",[M,F,length(A),A]),
	    From ! {apply, Res},
	    erlang:trace_pattern({?MODULE,slave,1},false,[meta]),
	    slave(From);
	{From, lambda, Fun} ->
	    Res = Fun(),
	    From ! {lambda, Res},
	    erlang:trace_pattern({?MODULE,slave,1},false,[meta]),
	    slave(From);
	die ->
	    exit(normal)
    end.

collect() ->
    L = lists:reverse(collect([], 0)),
    ?dbgformat("Got: ~p~n",[L]),
    L.

collect(Time) ->
    Ref = erlang:start_timer(Time, self(), done),
    L = lists:reverse(collect([], Ref)),
    ?dbgformat("Got: ~p~n",[L]),
    L.

collect(A, 0) ->
    receive
	Mess ->
	    collect([Mess | A], 0)
    after 0 ->
	    A
    end;
collect(A, Ref) ->
    receive
	{timeout, Ref, done} ->
	    collect(A, 0);
	Mess ->
	    collect([Mess | A], Ref)
    end.

setup() ->
    trace_off(),
    Self = self(),
    Pid = spawn(fun () -> slave(Self) end),
    receive sync -> ok end,
    put(slave,Pid),
    Pid.

shutdown() ->
    trace_off(),
    Pid = get(slave),
    case (catch is_process_alive(Pid)) of
	true ->
	    Ref = erlang:monitor(process,Pid),
	    Pid ! die,
	    receive
		{'DOWN',Ref,process,Pid,_} ->
		    ok
	    end;
	_ ->
	    ok
    end.

trace_off() ->
    erlang:trace(all, false, [all]),
    erlang:trace_pattern({'_','_','_'},false,[]),
    erlang:trace_pattern({'_','_','_'},false,[local]),
    erlang:trace_pattern({'_','_','_'},false,[meta]),
    erlang:trace_pattern(on_load,false,[]),
    erlang:trace_pattern(on_load,false,[local]),
    erlang:trace_pattern(on_load,false,[meta]),
    ok.

apply_slave_async(M,F,A) ->
    Slave = get(slave),
    Pid = 
	spawn(
	  fun () ->
		  Slave ! {self(),apply, M, F, A},
		  receive
		      {apply, _} ->
			  receive
			      sync ->
				  ok
			  end
		  end
	  end),
    Pid.

apply_slave(M,F,A) ->
    Pid = get(slave),
    Pid ! {self(),apply, M, F, A},
    receive
	{apply, Res} ->
	    receive
		sync ->
		    Res
	    end
    end.

lambda_slave(Fun) ->
    Pid = get(slave),
    Pid ! {self(), lambda, Fun},
    receive
	{lambda, Res} ->
	    receive
		sync ->
		    Res
	    end
    end.

relay_n(0, _) ->
    ok;
relay_n(N, Dest) ->
    receive Msg ->
	    Dest ! {self(), Msg},
	    relay_n(N-1, Dest)
    end.
